# This Build.PL for Mojolicious-Plugin-OpenTelemetry was generated by Dist::Zilla::Plugin::ModuleBuildTiny 0.017.
use strict;
use warnings;


use 5.013010;
use Module::Build::Tiny 0.034;
Build_PL();

